import os
import typing as TYPE


def typed_evar(name: str, default: TYPE.Optional[TYPE.Any] = None):
    """Return an environment variable with an assumed type. Type from
    the default value, if provided, will be prioritized, otherwise
    the type will be inferred in order of: bool, int, float, str.

    (CURDAY, 25) -> 25
    (CURDAY, 25.0) -> 25.0
    (CURDAY, '25.0') -> '25.0'
    (CURDAY, None) -> '25'

    Args:
        name (str): The name of the environment variable.
        default (Optional[Any]): The default value of the environment variable.
            Defaults to None.
    Returns:
        Any: The env var value with the assumed type.
    """
    varval = os.environ.get(name)
    if varval is None:
        return default

    # use default's type
    if default is not None:
        vartype = type(default)

        # bool gets special treatment
        if vartype is bool:
            if varval.lower() in ('1', 'true'):
                return True
            elif varval.lower() in ('0', 'false', '-1'):
                return False
            else:
                raise ValueError(
                    "Invalid boolean value for environment variable %s: %s"
                    % (name, varval)
                )
        try:
            return vartype(varval)
        except Exception:
            pass

    # otherwise assume type using a few simple types
    if varval.casefold() in ('true', 'false'):
        return varval.casefold() == 'true'

    for vartype in (int, float):
        try:
            return vartype(varval)
        except ValueError:
            continue

    return varval



class NotPrintableError(Exception):
    """Custom exception to be raised when both repr and str fail"""

    def __init__(self, obj, e_repr, e_str):
        self.obj_type = type(obj)
        self.repr_exception = e_repr
        self.str_exception = e_str
        self.message = (
            "<NotPrintable {obj_type} object repr_exception={e_repr!r} "
            "str_exception={e_str!r}>".format(
                obj_type=self.obj_type, e_repr=e_repr, e_str=e_str
            )
        )
        super().__init__(self.message)



from typing import Optional as Opt, Union, Any, Generator, Iterable
from random import choice, randint
from pathlib import Path
from .exceptions import NotPrintableError
from .consts import ALPHANUMERIC_CHARS, ALPHANUMERIC_EXT_CHARS


def ranstr(
    min_len: int = 16,
    max_len: Opt[int] = None,
    chars: Iterable = ALPHANUMERIC_CHARS,
    as_generator: bool = False,
) -> Union[Generator, str]:
    """Generates str with random chars between min and max length
    Args:
        min_length (int) = 16: The length of the string.
        max_length (Optional[int]): The maximum length of the string.
            If None, min_length is used with no variance on ranstr len
        chars (Iterable = ALPHANUMERIC_CHARS): Characters used for random str
        as_generator (bool) = False: Return a generator instead of a string
    Returns:
        Generator | str: The random str or generator for random str
    """
    str_len = min_len if max_len is None else randint(min_len, max_len)

    if as_generator:
        return (choice(chars) for _ in range(str_len))
    return ''.join(choice(chars) for _ in range(str_len))


def safe_repr(obj):
    """Try and returns an objects repr, then str if err, then NotPrintableError
    error message if both fail
    """
    try:
        return repr(obj)
    except Exception as e_repr:
        try:
            return str(obj)
        except Exception as e_str:
            err = NotPrintableError(obj, e_repr, e_str)
            return err.message


def default_repr(
    obj: Any,
    join_attrs_on: str = ', ',
    attrs_format: str = "{attr_name}={safe_attr_repr}",
    repr_format: str = "{obj_name}({attributes})",
) -> str:
    """Return a string representation of a custom Python object.
    This representation is constructed such that the object can be
    reconstructed from the returned string, ideally. Complex objects
    may not be able to be reconstructed.
    Args:
        obj (Any): The input Python object.
        attrs_join (str) = ', ': The string to join the attributes with.
        attrs_format (str) = "{attr_name}={safe_attr_repr}": The format string
            to use for each attribute.
    Returns:
        str: The string representation of the object.
    """
    # If the object has a __dict__ attribute, use that
    if hasattr(obj, '__dict__'):
        attributes = join_attrs_on.join(
            attrs_format.format(attr_name=key, safe_attr_repr=safe_repr(value))
            for key, value in obj.__dict__.items()
            if not hasattr(value, '__call__') and not str(key).startswith("_")
        )
    # Otherwise, use dir() to gather potential attributes
    else:
        if isinstance(obj, int):
            return '%d' % obj
        elif isinstance(obj, float):
            return '%f' % obj
        elif isinstance(obj, str):
            return '%s' % repr(obj)
        elif isinstance(obj, set):
            return '%s' % obj
        elif isinstance(obj, (list, tuple, dict)):
            return str(obj)

        attributes = join_attrs_on.join(
            attrs_format.format(
                attr_name=attr, safe_attr_repr=safe_repr(getattr(obj, attr))
            )
            for attr in dir(obj)
            if not hasattr(obj, '__call__') and not str(attr).startswith('_')
        )

    return repr_format.format(
        obj_name=obj.__class__.__name__, attributes=attributes
    )



from .consts import ALPHANUMERIC_CHARS, ALPHANUMERIC_EXT_CHARS
from .crypto import is_jwt
from .env import typed_evar
from .exceptions import NotPrintableError
from .python import default_repr, ranstr, safe_repr
from .shell import runcmd
from .terminal import get_terminal_width, print_columns, print_middle



import base64 as b64
from typing import Union
import json


def _pad(s: str) -> str:
    """
    Pad a string with '=' to make it base64 decodable.
    Args:
        s (str): Input string.
    Returns:
        str: The padded string.
    """
    padding = len(s) % 4
    return s if padding == 0 else s + '=' * (4 - padding)


def is_jwt(s: Union[str, bytes]) -> bool:
    """
    Check if a string is a valid JWT (JSON Web Token).
    Args:
        s (Union[str, bytes]): Input suspected JWT bytes or string.
    Returns:
        bool: True if the input is a valid JWT, False otherwise.
    """
    try:
        if isinstance(s, bytes):
            s = s.decode('utf-8')

        segments = s.split('.')
        if len(segments) != 3:
            return False

        # Decoding header and payload to check if they are valid JSON
        header, payload, signature = segments
        json.loads(b64.urlsafe_b64decode(_pad(header)).decode('utf-8'))
        json.loads(b64.urlsafe_b64decode(_pad(payload)).decode('utf-8'))

        # Rudimentary check for signature length
        if (
            len(signature) < 10
        ):  # Minimum length for a JWT signature is typically more than 10 characters
            return False

        return True
    except Exception:
        return False



import typing as TYPE
from shlex import split as shx_split
from subprocess import run as sp_run, CompletedProcess


def runcmd(
    cmd: TYPE.Union[str, TYPE.List], output: bool = True, *args, **kwargs
) -> TYPE.Optional[CompletedProcess]:
    """Runs a single command in the shell with subprocess.run
    Args:
        cmd (Union[str, List]): The command to run in the shell.
        output (bool): Whether or not to return the output of the command.
            Defaults to True.
    """
    if isinstance(cmd, str):
        cmd = shx_split(cmd)

    if output:
        return sp_run(
            cmd, check=True, text=True, capture_output=True, *args, **kwargs
        )
    else:
        sp_run(
            cmd, check=False, text=False, capture_output=False, *args, **kwargs
        )
